@@ -25,6 +25,8 @@ from subprocess import PIPE, STDOUT, Popen
 from twisted.python.versions import Version
 from twisted.python.filepath import FilePath
 from twisted.python.dist import twisted_subprojects
+from twisted.python.compat import execfile
+from twisted.python.usage import Options, UsageError
 
 # This import is an example of why you shouldn't use this module unless you're
 # radix
@@ -60,6 +62,7 @@ def runCommand(args):
     return stdout
 
 
+
 class CommandFailed(Exception):
     """
     Raised when a child process exits unsuccessfully.
@@ -90,27 +93,51 @@ def _changeVersionInFile(old, new, filename):
 
 
 
-def getNextVersion(version, now=None):
+def getNextVersion(version, prerelease, patch, today):
     """
     Calculate the version number for a new release of Twisted based on
     the previous version number.
 
     @param version: The previous version number.
-    @param now: (optional) The current date.
+
+    @type prerelease: C{bool}
+    @param prerelease: If C{True}, make the next version a pre-release one. If
+       C{version} is a pre-release, it increments the pre-release counter,
+       otherwise create a new version with prerelease set to 1.
+
+    @type patch: C{bool}
+    @param patch: If C{True}, make the next version a patch release. It
+        increments the micro counter.
+
+    @type today: C{datetime}
+    @param today: The current date.
     """
-    # XXX: This has no way of incrementing the patch number. Currently, we
-    # don't need it. See bug 2915. Jonathan Lange, 2007-11-20.
-    if now is None:
-        now = date.today()
-    major = now.year - VERSION_OFFSET
+    micro = 0
+    major = today.year - VERSION_OFFSET
     if major != version.major:
         minor = 0
     else:
         minor = version.minor + 1
-    return Version(version.package, major, minor, 0)
+
+    if patch:
+        micro = version.micro + 1
+        major = version.major
+        minor = version.minor
+
+    newPrerelease = None
+    if version.prerelease is not None:
+        major = version.major
+        minor = version.minor
+        micro = version.micro
+        if prerelease:
+            newPrerelease = version.prerelease + 1
+    elif prerelease:
+        newPrerelease = 1
+    return Version(version.package, major, minor, micro, newPrerelease)
 
 
-def changeAllProjectVersions(root, versionTemplate, today=None):
+
+def changeAllProjectVersions(root, prerelease, patch, today=None):
     """
     Change the version of all projects (including core and all subprojects).
 
@@ -119,34 +146,35 @@ def changeAllProjectVersions(root, versionTemplate, today=None):
 
     @type root: L{FilePath}
     @param root: The root of the Twisted source tree.
-    @type versionTemplate: L{Version}
-    @param versionTemplate: The version of all projects.  The name will be
-        replaced for each respective project.
-    @type today: C{str}
-    @param today: A YYYY-MM-DD formatted string. If not provided, defaults to
-        the current day, according to the system clock.
+
+    @type prerelease: C{bool}
+    @param prerelease:
+
+    @type patch: C{bool}
+    @param patch:
+
+    @type today: C{datetime}
+    @param today: Defaults to the current day, according to the system clock.
     """
     if not today:
-        today = date.today().strftime('%Y-%m-%d')
+        today = date.today()
+    formattedToday = today.strftime('%Y-%m-%d')
     for project in findTwistedProjects(root):
         if project.directory.basename() == "twisted":
             packageName = "twisted"
         else:
             packageName = "twisted." + project.directory.basename()
         oldVersion = project.getVersion()
-        newVersion = Version(packageName, versionTemplate.major,
-                             versionTemplate.minor, versionTemplate.micro,
-                             prerelease=versionTemplate.prerelease)
-
+        newVersion = getNextVersion(oldVersion, prerelease, patch, today)
         if oldVersion.prerelease:
             builder = NewsBuilder()
             builder._changeNewsVersion(
                 root.child("NEWS"), builder._getNewsName(project),
-                oldVersion, newVersion, today)
+                oldVersion, newVersion, formattedToday)
             builder._changeNewsVersion(
                 project.directory.child("topfiles").child("NEWS"),
                 builder._getNewsName(project), oldVersion, newVersion,
-                today)
+                formattedToday)
 
         # The placement of the top-level README with respect to other files (eg
         # _version.py) is sufficiently different from the others that we just
@@ -159,7 +187,6 @@ def changeAllProjectVersions(root, versionTemplate, today=None):
 
 
 
-
 class Project(object):
     """
     A representation of a project that has a version.
@@ -219,19 +246,6 @@ def findTwistedProjects(baseDirectory):
 
 
 
-def updateTwistedVersionInformation(baseDirectory, now):
-    """
-    Update the version information for Twisted and all subprojects to the
-    date-based version number.
-
-    @param baseDirectory: Where to look for Twisted. If None, the function
-        infers the information from C{twisted.__file__}.
-    @param now: The current date (as L{datetime.date}). If None, it defaults
-        to today.
-    """
-    for project in findTwistedProjects(baseDirectory):
-        project.updateVersion(getNextVersion(project.getVersion(), now=now))
-
 
 def generateVersionFileData(version):
     """
@@ -244,13 +258,23 @@ def generateVersionFileData(version):
     else:
         prerelease = ""
     data = '''\
+# Copyright (c) Twisted Matrix Laboratories.
+# See LICENSE for details.
+
 # This is an auto-generated file. Do not edit it.
+
+"""
+Provides Twisted version information.
+"""
+
 from twisted.python import versions
 version = versions.Version(%r, %s, %s, %s%s)
-''' % (version.package, version.major, version.minor, version.micro, prerelease)
+''' % (version.package, version.major, version.minor, version.micro,
+       prerelease)
     return data
 
 
+
 def replaceProjectVersion(filename, newversion):
     """
     Write version specification code into the given filename, which
@@ -272,17 +296,17 @@ def replaceInFile(filename, oldToNew):
     """
     I replace the text `oldstr' with `newstr' in `filename' using science.
     """
-    os.rename(filename, filename+'.bak')
-    f = open(filename+'.bak')
+    os.rename(filename, filename + '.bak')
+    f = open(filename + '.bak')
     d = f.read()
     f.close()
-    for k,v in oldToNew.items():
+    for k, v in oldToNew.items():
         d = d.replace(k, v)
     f = open(filename + '.new', 'w')
     f.write(d)
     f.close()
-    os.rename(filename+'.new', filename)
-    os.unlink(filename+'.bak')
+    os.rename(filename + '.new', filename)
+    os.unlink(filename + '.bak')
 
 
 
@@ -353,7 +377,8 @@ class DocBuilder(LoreBuilderMixin):
         inputFiles = docDir.globChildren("*.xhtml")
         filenames = [x.path for x in inputFiles]
         if not filenames:
-            raise NoDocumentsFound("No input documents found in %s" % (docDir,))
+            raise NoDocumentsFound(
+                "No input documents found in %s" % (docDir,))
         if apiBaseURL is not None:
             arguments = ["--config", "baseurl=" + apiBaseURL]
         else:
@@ -424,8 +449,7 @@ class APIBuilder(object):
     """
     Generate API documentation from source files using
     U{pydoctor<http://codespeak.net/~mwh/pydoctor/>}.  This requires
-    pydoctor to be installed and usable (which means you won't be able to
-    use it with Python 2.3).
+    pydoctor to be installed and usable.
     """
     def build(self, projectName, projectURL, sourceURL, packagePath,
               outputPath):
@@ -551,8 +575,8 @@ class BookBuilder(LoreBuilderMixin):
 
                 # What I tell you three times is true!
                 # The first two invocations of latex on the book file allows it
-                # correctly create page numbers for in-text references.  Why this is
-                # the case, I could not tell you. -exarkun
+                # correctly create page numbers for in-text references.  Why
+                # this is the case, I could not tell you. -exarkun
                 for i in range(3):
                     self.run(texToDVI)
 
@@ -634,8 +658,7 @@ class NewsBuilder(object):
         _BUGFIX: "Bugfixes",
         _DOC: "Improved Documentation",
         _REMOVAL: "Deprecations and Removals",
-        _MISC: "Other",
-        }
+        _MISC: "Other"}
 
     _NO_CHANGES = "No significant changes have been made for this release.\n"
 
@@ -671,8 +694,8 @@ class NewsBuilder(object):
             base, ext = os.path.splitext(child.basename())
             if ext == ticketType:
                 results.append((
-                        int(base),
-                        ' '.join(child.getContent().splitlines())))
+                    int(base),
+                    ' '.join(child.getContent().splitlines())))
         results.sort()
         return results
 
@@ -727,7 +750,7 @@ class NewsBuilder(object):
         fileObj.write(header + '\n' + '-' * len(header) + '\n')
         for (description, relatedTickets) in reverse:
             ticketList = ', '.join([
-                    '#' + str(ticket) for ticket in relatedTickets])
+                '#' + str(ticket) for ticket in relatedTickets])
             entry = ' - %s (%s)' % (description, ticketList)
             entry = textwrap.fill(entry, subsequent_indent='   ')
             fileObj.write(entry + '\n')
@@ -772,6 +795,8 @@ class NewsBuilder(object):
 
         @param header: The top-level header to use when writing the news.
         @type header: L{str}
+
+        @raise NotWorkingDirectory: If the C{path} is not an SVN checkout.
         """
         changes = []
         for part in (self._FEATURE, self._BUGFIX, self._DOC, self._REMOVAL):
@@ -800,6 +825,23 @@ class NewsBuilder(object):
         output.sibling('NEWS.new').moveTo(output)
 
 
+    def _deleteFragments(self, path):
+        """
+        Delete the change information, to clean up the repository  once the
+        NEWS files have been built. It requires C{path} to be in a SVN
+        directory.
+
+        @param path: A directory (probably a I{topfiles} directory) containing
+            change information in the form of <ticket>.<change type> files.
+        @type path: L{FilePath}
+        """
+        ticketTypes = self._headings.keys()
+        for child in path.children():
+            base, ext = os.path.splitext(child.basename())
+            if ext in ticketTypes:
+                runCommand(["svn", "rm", child.path])
+
+
     def _getNewsName(self, project):
         """
         Return the name of C{project} that should appear in NEWS.
@@ -818,11 +860,11 @@ class NewsBuilder(object):
         Iterate through the Twisted projects in C{baseDirectory}, yielding
         everything we need to know to build news for them.
 
-        Yields C{topfiles}, C{news}, C{name}, C{version} for each sub-project
-        in reverse-alphabetical order. C{topfile} is the L{FilePath} for the
-        topfiles directory, C{news} is the L{FilePath} for the NEWS file,
-        C{name} is the nice name of the project (as should appear in the NEWS
-        file), C{version} is the current version string for that project.
+        Yields C{topfiles}, C{name}, C{version}, for each sub-project in
+        reverse-alphabetical order. C{topfile} is the L{FilePath} for the
+        topfiles directory, C{name} is the nice name of the project (as should
+        appear in the NEWS file), C{version} is the current version string for
+        that project.
 
         @param baseDirectory: A L{FilePath} representing the root directory
             beneath which to find Twisted projects for which to generate
@@ -837,16 +879,11 @@ class NewsBuilder(object):
         # files.
         projects.reverse()
 
-        for aggregateNews in [False, True]:
-            for project in projects:
-                topfiles = project.directory.child("topfiles")
-                if aggregateNews:
-                    news = baseDirectory.child("NEWS")
-                else:
-                    news = topfiles.child("NEWS")
-                name = self._getNewsName(project)
-                version = project.getVersion()
-                yield topfiles, news, name, version
+        for project in projects:
+            topfiles = project.directory.child("topfiles")
+            name = self._getNewsName(project)
+            version = project.getVersion()
+            yield topfiles, name, version
 
 
     def buildAll(self, baseDirectory):
@@ -860,11 +897,24 @@ class NewsBuilder(object):
             beneath which to find Twisted projects for which to generate
             news (see L{findTwistedProjects}).
         """
+        try:
+            runCommand(["svn", "info", baseDirectory.path])
+        except CommandFailed:
+            raise NotWorkingDirectory(
+                "%s does not appear to be an SVN working directory."
+                % (baseDirectory.path,))
+
         today = self._today()
-        for topfiles, news, name, version in self._iterProjects(baseDirectory):
-            self.build(
-                topfiles, news,
-                "Twisted %s %s (%s)" % (name, version.base(), today))
+        for topfiles, name, version in self._iterProjects(baseDirectory):
+            # We first build for the subproject
+            news = topfiles.child("NEWS")
+            header = "Twisted %s %s (%s)" % (name, version.base(), today)
+            self.build(topfiles, news, header)
+            # Then for the global NEWS file
+            news = baseDirectory.child("NEWS")
+            self.build(topfiles, news, header)
+            # Finally, delete the fragments
+            self._deleteFragments(topfiles)
 
 
     def _changeNewsVersion(self, news, name, oldVersion, newVersion, today):
@@ -905,7 +955,8 @@ class NewsBuilder(object):
         @type args: C{list} of C{str}
         """
         if len(args) != 1:
-            sys.exit("Must specify one argument: the path to the Twisted checkout")
+            sys.exit("Must specify one argument: the path to the "
+                     "Twisted checkout")
         self.buildAll(FilePath(args[0]))
 
 
@@ -998,8 +1049,9 @@ class DistributionBuilder(object):
             self.manBuilder.build(path)
         if path.isdir():
             try:
-                self.docBuilder.build(version, howtoPath, path,
-                    self.templatePath, self.apiBaseURL, True)
+                self.docBuilder.build(
+                    version, howtoPath, path, self.templatePath,
+                    self.apiBaseURL, True)
             except NoDocumentsFound:
                 pass
 
@@ -1029,8 +1081,8 @@ class DistributionBuilder(object):
             for subProjectDir in docPath.children():
                 if subProjectDir.isdir():
                     for child in subProjectDir.walk():
-                        self._buildDocInDir(child, version,
-                            subProjectDir.child("howto"))
+                        self._buildDocInDir(
+                            child, version, subProjectDir.child("howto"))
 
         for binthing in self.rootDirectory.child("bin").children():
             # bin/admin should not be included.
@@ -1048,7 +1100,8 @@ class DistributionBuilder(object):
                                                       submodule.basename()))
 
         for docDir in self.rootDirectory.child("doc").children():
-            tarball.add(docDir.path, buildPath("doc", docDir.basename()))
+            if docDir.basename() != "historic":
+                tarball.add(docDir.path, buildPath("doc", docDir.basename()))
 
         for toplevel in self.rootDirectory.children():
             if not toplevel.isdir():
@@ -1093,7 +1146,8 @@ class DistributionBuilder(object):
             if path.basename() == "plugins":
                 for plugin in path.children():
                     for subproject in self.subprojects:
-                        if plugin.basename() == "twisted_%s.py" % (subproject,):
+                        if (plugin.basename() ==
+                                "twisted_%s.py" % (subproject,)):
                             break
                     else:
                         tarball.add(plugin.path,
@@ -1244,7 +1298,7 @@ def buildAllTarballs(checkout, destination, templatePath=None):
     if not destination.exists():
         destination.createDirectory()
     db = DistributionBuilder(export, destination, templatePath=templatePath,
-        apiBaseURL=apiBaseURL)
+                             apiBaseURL=apiBaseURL)
 
     db.buildCore(versionString)
     for subproject in twisted_subprojects:
@@ -1255,6 +1309,16 @@ def buildAllTarballs(checkout, destination, templatePath=None):
     workPath.remove()
 
 
+
+class ChangeVersionsScriptOptions(Options):
+    """
+    Options for L{ChangeVersionsScript}.
+    """
+    optFlags = [["prerelease", None, "Change to the next prerelease"],
+                ["patch", None, "Make a patch version"]]
+
+
+
 class ChangeVersionsScript(object):
     """
     A thing for changing version numbers. See L{main}.
@@ -1270,31 +1334,15 @@ class ChangeVersionsScript(object):
         @param args: List of command line arguments.  This should only
             contain the version number.
         """
-        version_format = (
-            "Version should be in a form kind of like '1.2.3[pre4]'")
-        if len(args) != 1:
-            sys.exit("Must specify exactly one argument to change-versions")
-        version = args[0]
+        options = ChangeVersionsScriptOptions()
+
         try:
-            major, minor, micro_and_pre = version.split(".")
-        except ValueError:
-            raise SystemExit(version_format)
-        if "pre" in micro_and_pre:
-            micro, pre = micro_and_pre.split("pre")
-        else:
-            micro = micro_and_pre
-            pre = None
-        try:
-            major = int(major)
-            minor = int(minor)
-            micro = int(micro)
-            if pre is not None:
-                pre = int(pre)
-        except ValueError:
-            raise SystemExit(version_format)
-        version_template = Version("Whatever",
-                                   major, minor, micro, prerelease=pre)
-        self.changeAllProjectVersions(FilePath("."), version_template)
+            options.parseOptions(args)
+        except UsageError as e:
+            raise SystemExit(e)
+
+        self.changeAllProjectVersions(FilePath("."), options["prerelease"],
+                                      options["patch"])
 
 
 
@@ -1311,15 +1359,15 @@ class BuildTarballsScript(object):
         @type args: list of C{str}
         @param args: The command line arguments to process.  This must contain
             at least two strings: the checkout directory and the destination
-            directory. An optional third string can be specified for the website
-            template file, used for building the howto documentation. If this
-            string isn't specified, the default template included in twisted
-            will be used.
+            directory. An optional third string can be specified for the
+            website template file, used for building the howto documentation.
+            If this string isn't specified, the default template included in
+            twisted will be used.
         """
         if len(args) < 2 or len(args) > 3:
             sys.exit("Must specify at least two arguments: "
-                     "Twisted checkout and destination path. The optional third "
-                     "argument is the website template path.")
+                     "Twisted checkout and destination path. The optional "
+                     "third argument is the website template path.")
         if len(args) == 2:
             self.buildAllTarballs(FilePath(args[0]), FilePath(args[1]))
         elif len(args) == 3:
@@ -1367,3 +1415,69 @@ class BuildAPIDocsScript(object):
             sys.exit("Must specify two arguments: "
                      "Twisted checkout and destination path")
         self.buildAPIDocs(FilePath(args[0]), FilePath(args[1]))
+
+
+
+class BuildDocsScript(object):
+    """
+    A thing for building the main documentation. See L{main}.
+    """
+
+    def buildDocs(self, projectRoot, template):
+        """
+        Build the main documentation of Twisted, with our project policy.
+
+        @param projectRoot: A L{FilePath} representing the root of the Twisted
+            checkout.
+        @type projectRoot: L{FilePath}
+
+        @param template: A L{FilePath} pointing to the template file that is
+            used for the look and feel of the howto documentation.
+        @type template: L{FilePath}
+        """
+        version = Project(projectRoot.child("twisted")).getVersion()
+        versionString = version.base()
+        apiURL = "http://twistedmatrix.com/documents/%s/api/" % (
+            versionString,)
+
+        dirs = {}
+        docRoot = projectRoot.child("doc")
+
+        for p in docRoot.walk():
+            if p.basename() == 'man':
+                ManBuilder().build(p)
+
+        howtoRoot = docRoot.descendant(["core", "howto"])
+
+        for p in docRoot.walk():
+            if p.basename().endswith('.xhtml'):
+                if p.parent() not in dirs:
+                    dirs[p.parent()] = True
+                    DocBuilder().build(
+                        versionString, howtoRoot, p.parent(),
+                        template,
+                        apiURL + "%s.html",
+                        False)
+
+        BookBuilder().build(howtoRoot, dirs.keys(),
+                            howtoRoot.child('book.tex'),
+                            howtoRoot.child('book.pdf'))
+
+        for path in dirs:
+            for doc in path.globChildren("*.xhtml"):
+                doc.remove()
+
+
+    def main(self, args):
+        """
+        Build the main documentation.
+
+        @type args: list of str
+        @param args: The command line arguments to process.  This must contain
+            two strings: the path to the root of the Twisted checkout and the
+            path to the Twisted website template.
+        """
+        if len(args) != 2:
+            sys.exit("Must specify two arguments: "
+                     "Twisted checkout path and template path.")
+        self.buildDocs(FilePath(args[0]), FilePath(args[1]))
